package door

import (
	"bufio"
	"bytes"
	"io"
	"log"

	"syscall"
	"unicode"

	"github.com/creack/goselect"
)

const READ_SIZE = 16 // Size of read buffer

var readerBuffer []rune

/* // linux Reader
// go routine Reader for input
// This "times out" every ReaderTimeval
func Reader(d *Door) {
	// I need non-blocking reads here.

	readerBuffer = make([]rune, 0, READ_SIZE*2)

	defer func() {
		log.Printf("~Reader\n")
		if d.ReaderCanClose {
			d.wg.Done()
		}
	}()
	defer func() {
		if err := recover(); err != nil {
			log.Printf("Reader: %#v\n", err)
		}
	}()

	var fdset syscall.FdSet
	var readbuffer [READ_SIZE]byte
	var runebuffer bytes.Buffer
	var runeread = bufio.NewReaderSize(&runebuffer, 1)

	for {
		clearAll(&fdset)
		set(&fdset, d.READFD)

		v, err := syscall.Select(d.READFD+1, &fdset, nil, nil, &ReaderTimeval)

		if err == syscall.EINTR {
			continue
		}
		// log.Printf("Select: %#v / %#v\n", v, err)

		if v == -1 {
			log.Printf("Reader ERR: %#v\n", err)
			d.readerMutex.Lock()

			if !d.ReaderClosed {
				d.ReaderClosed = true
				close(d.readerChannel)
			}
			return
		}

		if v == 0 {
			// timeout
			process(d, false)
			continue
		}

		// The buffer used here must have len & cap to size you want to read.
		// use [BUFF_SIZE]byte for readone
		r, err := syscall.Read(d.READFD, readbuffer[:])
		if r == -1 {
			log.Println("syscall.Read -1 (closed)")
			d.readerMutex.Lock()
			defer d.readerMutex.Unlock()
			if !d.ReaderClosed {
				d.ReaderClosed = true
				d.Disconnected = true
				close(d.readerChannel)
			}
			return
		}
		if r == 0 {
			log.Printf("Select said ready, but: %#v %#v\n", r, err)
			d.readerMutex.Lock()
			defer d.readerMutex.Unlock()
			if !d.ReaderClosed {
				d.ReaderClosed = true
				d.Disconnected = true
				close(d.readerChannel)
			}
			return
		}

		if DEBUG_INPUT {
			log.Printf("Reader << %d, %#v\n", r, readbuffer[:r])
		}

		runebuffer.Write(readbuffer[:r])
		var input rune

		// Is this unicode?

		err = nil
		for err == nil {
			//RuneRead:

			input, _, err = runeread.ReadRune()

			if err == io.EOF {
				break // continue
			}
			if err != nil {
				log.Printf("ReadRune: %#v\n", err)
				// errors EOF
				break // continue // break for loop
			}
			if input == unicode.ReplacementChar {
				runeread.UnreadRune()
				b, _ := runeread.ReadByte()
				if DEBUG_INPUT {
					log.Printf("Reader (byte) >> %x\n", b)
				}
				readerBuffer = append(readerBuffer, rune(b))
				// d.readerChannel <- rune(b)
			} else {
				if DEBUG_INPUT {
					log.Printf("Reader >> %x\n", input)
				}
				readerBuffer = append(readerBuffer, input)
				// d.readerChannel <- input
			}
			// process(d, true)

		} // goto RuneRead

		process(d, true)

		// buffer = append(buffer, readone[0])

		// d.readerChannel <- rune(buffer[0])

	}
}
*/

func Reader(d *Door) {
	// handle syscall.Handle, readerChannel *chan byte
	var handle syscall.Handle = syscall.Handle(d.READFD) // d.Config.Comm_handle)

	defer func() {
		log.Printf("~Reader")
		if d.ReaderCanClose {
			d.wg.Done()
		}
	}()

	defer func() {
		if err := recover(); err != nil {
			log.Printf("Reader: %#v\n", err)
		}
	}()

	var fdset *goselect.FDSet = &goselect.FDSet{}
	var readbuffer [READ_SIZE]byte
	var runebuffer bytes.Buffer
	var runeread = bufio.NewReaderSize(&runebuffer, 1)
	WSA_Buffer := syscall.WSABuf{Len: READ_SIZE, Buf: &readbuffer[0]}

	for {
		fdset.Zero()
		fdset.Set(uintptr(d.READFD))

		err := goselect.Select(d.READFD+1, fdset, nil, nil, ReaderInterval)

		if err == syscall.EINTR {
			continue
		}

		if err != nil {
			log.Printf("Reader Err: %#v\n", err)
			d.readerMutex.Lock()
			defer d.readerMutex.Unlock()
			if !d.ReaderClosed {
				d.ReaderClosed = true
				close(d.readerChannel)
			}
			return

		}

		if !fdset.IsSet(uintptr(d.READFD)) {
			// I believe this means timeout
			process(d, false)
			continue
		}

		read := uint32(0)
		flags := uint32(0)

		err = syscall.WSARecv(handle, &WSA_Buffer, 1, &read, &flags, nil, nil)
		if err != nil {
			log.Printf("Reader ERR: %#v\n", err)
			d.readerMutex.Lock()
			defer d.readerMutex.Unlock()
			if !d.ReaderClosed {
				d.ReaderClosed = true
				close(d.readerChannel)
			}
			return
		}

		if read <= 0 {
			log.Printf("READ FAILED %d\n", read)
			d.readerMutex.Lock()
			defer d.readerMutex.Unlock()
			if !d.ReaderClosed {
				d.ReaderClosed = true
				d.Disconnected = true
				close(d.readerChannel)
				return
			}
		}

		if DEBUG_INPUT {
			log.Printf("Reader << %d, %#v\n", read, readbuffer[:read])
		}

		runebuffer.Write(readbuffer[:read])
		var input rune
		err = nil
		for err == nil {
			input, _, err = runeread.ReadRune()

			if err == io.EOF {
				break // continue
			}
			if err != nil {
				log.Printf("ReadRune: %#v\n", err)
				// errors EOF
				break // continue // break for loop
			}
			if input == unicode.ReplacementChar {
				runeread.UnreadRune()
				b, _ := runeread.ReadByte()
				if DEBUG_INPUT {
					log.Printf("Reader (byte) >> %x\n", b)
				}
				readerBuffer = append(readerBuffer, rune(b))
				// d.readerChannel <- rune(b)
			} else {
				if DEBUG_INPUT {
					log.Printf("Reader >> %x\n", input)
				}
				readerBuffer = append(readerBuffer, input)
			}
		}

		process(d, true)
	}
}
